#ifndef __NODE_0_MODULE__
#define __NODE_0_MODULE__

#include <string>
#include <systemc>
using namespace sc_core;

class Node_0_module : public sc_module
{
private:
    uint64_t no_ins, no_outs;
    void set_sysc_output();
public:
    uint64_t no_inputs, no_outputs;
    sc_fifo<sc_dt::sc_bv<32>>* sysc_output;
    int get_no_inputs() const;
    int get_no_outputs() const;
    void body();
    SC_HAS_PROCESS(Node_0_module);
    Node_0_module(sc_module_name name);
};
#endif
#ifndef __NODE_1_MODULE__
#define __NODE_1_MODULE__

#include <string>
#include <systemc>
using namespace sc_core;

class Node_1_module : public sc_module
{
private:
    uint64_t no_ins, no_outs;
    void set_sysc_output();
public:
    uint64_t no_inputs, no_outputs;
    sc_fifo<sc_dt::sc_bv<32>>* sysc_output;
    int get_no_inputs() const;
    int get_no_outputs() const;
    void body();
    SC_HAS_PROCESS(Node_1_module);
    Node_1_module(sc_module_name name);
};
#endif
#ifndef __PRINT_THEN_EXIT_3_MODULE__
#define __PRINT_THEN_EXIT_3_MODULE__

#include <string>
#include <systemc>
using namespace sc_core;
#include "Python.h"

class Print_Then_Exit_3_module : public sc_module
{
private:
    PyObject *pBody, *pName, *pModule, *pyC, *pExit;
    uint64_t no_ins, no_outs;
    PyObject* type_sysc_n;
    PyObject* get_sysc_n();
    sc_dt::sc_bv<32> bits_sysc_n;
public:
    uint64_t no_inputs, no_outputs;
    sc_fifo<sc_dt::sc_bv<32>>* sysc_n;
    int get_no_inputs() const;
    int get_no_outputs() const;
    void body();
    SC_HAS_PROCESS(Print_Then_Exit_3_module);
    Print_Then_Exit_3_module(sc_module_name name);
};
#endif

#ifndef __TEST_PYTHON_TEMPLATE__
#define __TEST_PYTHON_TEMPLATE__
#include <systemc>
using namespace sc_core;
using namespace sc_dt;

// TODO: Find better way of handling trace files
sc_trace_file *Tf;

// Converts clock signals to bit vectors for Migen nodes
SC_MODULE(ClkToBV) {
    sc_in<bool> clk;
    sc_out<sc_bv<1>> clkout;

    void run() {
        clkout.write(clk.read());
    }

    SC_CTOR(ClkToBV) {
        SC_METHOD(run);
        sensitive << clk;
    }
};

// Adaptor for going from Python to Migen
template <class T> SC_MODULE(PythonToMigen) {
    sc_in<bool> clk;
    sc_out<T> migen_data_out;
    sc_out<sc_bv<1>> migen_valid_out;
    sc_in<sc_bv<1>> migen_ready_in;
    sc_fifo<T>* py_in;

    void run() {
        if (migen_ready_in.read() == 1) {
            T val;
            if (py_in->nb_read(val)) {
                migen_data_out.write(val);
                migen_valid_out.write(1);
            } else {
                migen_valid_out.write(0);
            }
        }
    }

    SC_CTOR(PythonToMigen) {
        SC_METHOD(run);
        sensitive << clk.pos();
    }
};

// Adaptor for going from Migen to Python
template <class T> SC_MODULE(MigenToPython) {
    sc_in<bool> clk;
    sc_in<T> migen_in;
    sc_in<sc_bv<1>> migen_valid_in;
    sc_out<sc_bv<1>> migen_ready_out;
    sc_fifo<T>* py_out;

    void run() {
        while (true) {
            wait();
            if (py_out->num_free() > 0) {
                migen_ready_out.write(1);
                if (migen_valid_in.read() == 1) {
                    py_out->write(migen_in.read());
                }
            } else {
                migen_ready_out.write(0);
            }
        }
    }

    SC_CTOR(MigenToPython) {
        SC_THREAD(run);
        sensitive << clk.pos();
    }
};

SC_MODULE(Test_Python_Template){

// Python nodes to Python nodes just need a queue

// Migen nodes to Migen nodes need data wires,
// as well as valid and ready signals

// Python to Migen need an adaptor which maps a queue
// to data, valid and ready signals.

// Migen to Python need an adaptor which maps signals
// to a queue

// Python to template need a public queue
sc_fifo<sc_dt::sc_bv<32>> node_0_sysc_output;
sc_fifo<sc_dt::sc_bv<32>> node_1_sysc_output;

// Template to python need a public queue
sc_fifo<sc_dt::sc_bv<32>> print_then_exit_3_sysc_n;

// Migen to template need public data, valid and ready signals

// Template to Migen need public data, valid and ready signals

sc_in<bool> clk, rst;
sc_signal<sc_dt::sc_bv<1>> rst_bv;

// Node modules
Node_0_module node_0;
Node_1_module node_1;
Print_Then_Exit_3_module print_then_exit_3;


// Constructor first initialises queues, modules and adaptors
typedef Test_Python_Template SC_CURRENT_USER_MODULE;
Test_Python_Template(sc_module_name name, sc_trace_file *Tf):
node_0("node_0"),
node_1("node_1"),
print_then_exit_3("print_then_exit_3")
{
    SC_METHOD(rstprop);
    sensitive << rst;

    // Wiring the clock to the Migen nodes

    // Wiring the Python to Python nodes

    // Wiring the Migen to Migen nodes

    // Wiring the Python to Migen nodes

    // Wiring the Migen to Python nodes

    // Wiring the Python to template nodes
    node_0.sysc_output = &node_0_sysc_output;
    node_1.sysc_output = &node_1_sysc_output;

    // Wiring the template to Python nodes
    print_then_exit_3.sysc_n = &print_then_exit_3_sysc_n;

    // Wiring the Migen to template nodes

    // Wiring the template to Migen nodes

    // Add tracing




    node_0_sysc_output.trace(Tf);
    node_1_sysc_output.trace(Tf);

    print_then_exit_3_sysc_n.trace(Tf);


}

void rstprop() {
    // Propogate reset signal to Migen nodes
    rst_bv.write(rst.read());
}

};

#endif
